use chrono::offset::Utc;
use curl::{easy, easy::Easy2};
use rand::prelude::*;
use rusqlite::{types::ToSqlOutput, Connection, ToSql};
use tauri::{Manager, State};
use tokio::sync::Mutex;

struct Collector(Vec<u8>);

impl easy::Handler for Collector {
	fn write(&mut self, data: &[u8]) -> Result<usize, easy::WriteError> {
		self.0.extend_from_slice(data);
		Ok(data.len())
	}
}

struct AppState {
	db: Connection,
	acc: i64,
}

#[derive(Clone, Copy)]
enum Store {
	Aldi,
	Edeka,
	Dm,
	Lidl,
	Rewe,
	Tegut,
}

impl TryFrom<&str> for Store {
	type Error = ();

	fn try_from(s: &str) -> Result<Self, Self::Error> {
		match s {
			"aldi" => Ok(Store::Aldi),
			"edeka" => Ok(Store::Edeka),
			"dm" => Ok(Store::Dm),
			"lidl" => Ok(Store::Lidl),
			"rewe" => Ok(Store::Rewe),
			"tegut" => Ok(Store::Tegut),
			_ => Err(()),
		}
	}
}

impl ToSql for Store {
	fn to_sql(&self) -> rusqlite::Result<ToSqlOutput<'_>> {
		match self {
			Store::Aldi => 0.to_sql(),
			Store::Edeka => 1.to_sql(),
			Store::Dm => 2.to_sql(),
			Store::Lidl => 3.to_sql(),
			Store::Rewe => 4.to_sql(),
			Store::Tegut => 5.to_sql(),
		}
	}
}

#[tauri::command]
async fn pull_data() -> Result<String, ()> {
	let mut client = Easy2::new(Collector(Vec::new()));
	client.custom_request("PROPFIND").map_err(|_| ())?;
	client.url("https://cloud.seebruecke.org/public.php/webdav/data/")
		.map_err(|_| ())?;
	client.username(include_str!("cloud_user.txt"));
	client.http_auth(easy::Auth::new().auto(true));
	client.perform().map_err(|_| ())?;
	let content = client.get_ref();
	Ok(String::from_utf8_lossy(&content.0).to_string())
}

#[tauri::command]
async fn push_data() -> Result<(), ()> {
	let mut client = Easy2::new(Collector(Vec::new()));
	Ok(())
}

// Learn more about Tauri commands at https://tauri.app/develop/calling-rust/
#[tauri::command]
async fn swap(
	store: &str,
	state: State<'_, Mutex<AppState>>,
) -> Result<(), ()> {
	let state = state.lock().await;
	let mut rng = rand::rng();
	let store: Store = store.try_into()?;
	state.db.execute(
		"INSERT INTO swap VALUES (?1, ?2, ?3, ?4, ?5)",
		(
			rng.random::<i64>(),
			store,
			state.acc,
			Utc::now().timestamp(),
			false,
		),
	)
	.map_err(|_| ())?;
	Ok(())
}

#[tauri::command]
async fn count(state: State<'_, Mutex<AppState>>) -> Result<String, ()> {
	let state = state.lock().await;
	let mut stmt = state.db.prepare("SELECT COUNT(*) FROM swap").unwrap();
	let mut rows = stmt.query([]).unwrap();
	let row = rows.next().unwrap().unwrap();
	let cnt: u64 = row.get_unwrap(0);
	Ok(cnt.to_string())
}

#[cfg_attr(mobile, tauri::mobile_entry_point)]
pub fn run() {
	let mut rng = rand::rng();
	let db = Connection::open_in_memory().expect("Failed to create DB.");
	let acc = rng.random::<i64>();
	db.execute(
		"CREATE TABLE swap (
			rand INTEGER,
			store INTEGER,
			account INTEGER,
			time INTEGER,
			cancellation BOOL
		)",
		(),
	)
	.unwrap();
	let state = AppState { db, acc };
	tauri::Builder::default()
		.plugin(tauri_plugin_fs::init())
		.setup(|app| {
			app.manage(Mutex::new(state));
			Ok(())
		})
		.plugin(tauri_plugin_opener::init())
		.invoke_handler(tauri::generate_handler![
			swap, count, pull_data
		])
		.run(tauri::generate_context!())
		.expect("error while running tauri application");
}
